#include <stdbool.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "waveforms.h"
#include "notes.h"
#include "synth.h"

#define MIDI_BAUD (31250UL)

volatile uint8_t envelope_advance = 0;

synth_patch brass = {
	.operators = {
		{
			.envelope = {
				.value = 0,
				.attack = 400,
				.decay = 100,
				.sustain = 96 << 8,
				.release = 30
			},
			.frequency_scale = 1024,
			.amplitude = 256
		},
		{
			.envelope = {
				.value = 0,
				.attack = 20,
				.decay = 5,
				.sustain = 240 << 8,
				.release = 40,
			},
			.frequency_scale = 515,
			.amplitude = 900
		}
	}
};

synth_patch flute = {
	.operators = {
		{
			.envelope = {
				.value = 0,
				.attack = 118,
				.decay = 23,
				.sustain = 128 << 8,
				.release = 23
			},
			.frequency_scale = 2068,
			.amplitude = 256
		},
		{
			.envelope = {
				.value = 0,
				.attack = 23,
				.decay = 30,
				.sustain = 64 << 8,
				.release = 40,
			},
			.frequency_scale = 1024,
			.amplitude = 100
		}	
	}
};

synth_patch e_piano = {
	.operators = {
		{
			.envelope = {
				.value = 0,
				.attack = 5292,
				.decay = 25,
				.sustain = 64 << 8,
				.release = 5
			},
			.frequency_scale = 3072,
			.amplitude = 256
		},
		{
			.envelope = {
				.value = 0,
				.attack = 5292,
				.decay = 49,
				.sustain = 128 << 8,
				.release = 81,
			},
			.frequency_scale = 1024,
			.amplitude = 100
		}	
	}
};

synth_patch organ = {
	.operators = {
		{
			.envelope = {
				.value = 0,
				.attack = 3200,
				.decay = 10,
				.sustain = 200 << 8,
				.release = 48
			},
			.frequency_scale = 1024,
			.amplitude = 256
		},
		{
			.oscillator = {
				.shape = SYNTH_WAVEFORM_SHAPE_TRIANGLE
			},
			.envelope = {
				.value = 0,
				.attack = 1200,
				.decay = 200,
				.sustain = 60 << 8,
				.release = 48,
			},
			.frequency_scale = 512,
			.amplitude = 300
		}	
	}
};

synth_patch a_guitar = {
	.operators = {
		{
			.envelope = {
				.value = 0,
				.attack = 65535,
				.decay = 40,
				.sustain = 0,
				.release = 20
			},
			.frequency_scale = 2048,
			.amplitude = 256
		},
		{
			.oscillator = {
				.shape = SYNTH_WAVEFORM_SHAPE_TRIANGLE
			},
			.envelope = {
				.value = 0,
				.attack = 600,
				.decay = 100,
				.sustain = 200 << 8,
				.release = 48,
			},
			.frequency_scale = 512,
			.amplitude = 256
		}	
	}
};

synth_patch e_guitar = {
	.operators = {
		{
			.oscillator = {
				.shape = SYNTH_WAVEFORM_SHAPE_TRIANGLE
			},
			.envelope = {
				.value = 0,
				.attack = 65535,
				.decay = 18,
				.sustain = 0,
				.release = 22
			},
			.frequency_scale = 1536,
			.amplitude = 256
		},
		{
			.oscillator = {
				.shape = SYNTH_WAVEFORM_SHAPE_TRIANGLE
			},
			.envelope = {
				.value = 0,
				.attack = 600,
				.decay = 5000,
				.sustain = 200 << 8,
				.release = 48,
			},
			.frequency_scale = 512,
			.amplitude = 300
		}	
	}
};


volatile uint8_t midi_status;
volatile uint8_t midi_data[2];
volatile uint8_t midi_data_index;
volatile uint8_t midi_data_bytes_per_message;

#define MIDI_QUEUE_LENGTH (128)
volatile uint8_t midi_queue[3 * MIDI_QUEUE_LENGTH];
volatile uint8_t midi_queue_read = 0;
volatile uint8_t midi_queue_write = 0;
volatile uint8_t midi_queue_count = 0;

ISR(USART_RX_vect) {
	uint8_t data = UDR0;
	if (data >= 0xF8) {
		// system control message, ignore
	} else if (data >= 0x80) {
		midi_status = data;
		midi_data_index = 0;
		midi_data[0] = 0;
		midi_data[1] = 0;
		midi_data_bytes_per_message = 2;
		switch (midi_status & 0xF0) {
			case 0xC0: // program change
			case 0xD0: // channel pressure (after-touch)
				midi_data_bytes_per_message = 1;
				break;
			case 0xF0: // sysex, time code, song position, song select, tune request, end of exclusive
				midi_data_bytes_per_message = 0;
				break;
		}
	} else if (midi_data_bytes_per_message > 0) {
		
		midi_data[midi_data_index++] = data;
		
		if (midi_data_index == midi_data_bytes_per_message) {
			
			if (midi_queue_count < MIDI_QUEUE_LENGTH) {
				midi_queue[midi_queue_write++] = midi_status;
				midi_queue[midi_queue_write++] = midi_data[0];
				midi_queue[midi_queue_write++] = midi_data[1];
				
				if (midi_queue_write >= 3 * MIDI_QUEUE_LENGTH) {
					midi_queue_write = 0;
				}
				++midi_queue_count;
			}
			
			midi_data_index = 0;
			midi_data[0] = 0;
			midi_data[1] = 0;
		}
	}
}



void note_off(uint8_t channel, uint8_t note) {
	
	for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
		if (synth_voices[v].channel == channel && synth_voices[v].note == note) {
			cli();
			synth_end_voice(&synth_voices[v]);
			sei();
		}
	}
	
}


void note_on(uint8_t channel, uint8_t note, uint8_t velocity) {
	
	note_off(channel, note);
	
	uint8_t voice_slot = 0xFF;
	
	// If there's a free slot, use that.
	for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
		if (synth_voices[v].state == SYNTH_VOICE_STATE_IDLE) {
			voice_slot = v;
			break;
		}
	}
	
	// If there's a matching note and channel use that.
	if (voice_slot >= SYNTH_VOICE_COUNT) {
		for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
			if (synth_voices[v].channel == channel && synth_voices[v].note == note) {
				voice_slot = v;
				break;
			}
		}
	}
	
	// Replace the ending voice.
	if (voice_slot >= SYNTH_VOICE_COUNT) {
		uint8_t oldest_voice = 0;
		for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
			if (synth_voices[v].age >= oldest_voice && synth_voices[v].state != SYNTH_VOICE_STATE_ACTIVE) {
				voice_slot = v;
				oldest_voice = synth_voices[v].age;
			}
		}
	}
	
	// Replace the oldest voice.
	if (voice_slot >= SYNTH_VOICE_COUNT) {
		uint8_t oldest_voice = 0;
		for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
			if (synth_voices[v].age >= oldest_voice) {
				voice_slot = v;
				oldest_voice = synth_voices[v].age;
			}
		}
	}
	
	
	if (voice_slot < SYNTH_VOICE_COUNT) {
		cli();
		synth_begin_voice(&synth_voices[voice_slot], channel, note, velocity);
		sei();
		for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
			if (v == voice_slot) {
				synth_voices[v].age = 0;
			} else if (synth_voices[v].age < 255) {
				++synth_voices[v].age;
			}
		}
	}
	
}

void pitch_bend(uint8_t channel, int16_t amount) {
	synth_set_pitch_bend(channel, amount);
}

void change_program(uint8_t channel, uint8_t program) {
	switch (program) {
		case 0:
			synth_set_patch(channel, &brass);
			break;
		case 1:
			synth_set_patch(channel, &flute);
			break;
		case 2:
			synth_set_patch(channel, &e_piano);
			break;
		case 3:
			synth_set_patch(channel, &organ);
			break;
		case 4:
			synth_set_patch(channel, &a_guitar);
			break;
		case 5:
			synth_set_patch(channel, &e_guitar);
			break;
	}
}


__attribute__((noreturn)) void main() {
	
	PORTC |= _BV(5);
	
	// We'll use fast PWM mode to generate an output analogue waveform.
	TCCR0A = (1 << COM0A1) | (1 << COM0A0) | (1 << COM0B1) | (1 << COM0B0) | (1 << WGM01) | (1 << WGM00);
	TCCR0B = (0 << WGM02) | (1 << CS00); // No prescaler
	DDRD |= _BV(6) | _BV(5);

	// Set up timer to run at the output sample rate.
	TCCR1B |= (0b001 << CS10); // Run at native clock speed, no prescaling.
	TIMSK1 |= _BV(OCIE1A);     // Interrupt on match A.
	TCCR1B |= _BV(WGM12);      // Mode 4 (TOP=OCR1A, CTC).
	OCR1A = (F_CPU / SYNTH_SAMPLE_RATE) - 1;
	
	// Set up the USART to receive MIDI
	UCSR0B = (1 << RXCIE0) | (1 << RXEN0) | (0 << UCSZ02);
	UCSR0C = (1 << UCSZ01) | (1 << UCSZ00); // 8 bits per character
	UBRR0 = (F_CPU / 16 / MIDI_BAUD) - 1;
	
	synth_set_patch(0, &brass);
	synth_set_patch(1, &flute);
	synth_set_patch(2, &e_piano);
	synth_set_patch(3, &organ);
	synth_set_patch(4, &a_guitar);
	synth_set_patch(5, &e_guitar);
	
	sei();
	
	for (;;) {
		
		while (midi_queue_count > 0) {
			uint8_t status = midi_queue[midi_queue_read++];
			uint8_t data0 = midi_queue[midi_queue_read++];
			uint8_t data1 = midi_queue[midi_queue_read++];
			if (midi_queue_read >= 3 * MIDI_QUEUE_LENGTH) {
				midi_queue_read = 0;
			}
			--midi_queue_count;
			uint8_t channel = (status & 0x0F);
			switch (status & 0xF0) {
				case 0x80: // note off, note, velocity
					note_off(channel, data0);
					break;
				case 0x90: // note on, note, velocity
					if (data1 == 0) {
						note_off(channel, data0);
					} else {
						note_on(channel, data0, data1);
					}
					break;
				case 0xA0: // poly pressure (aftertouch), note, pressure
					break;
				case 0xB0: // control change, control, value
					break;
				case 0xC0: // program change, program
					change_program(channel, data0);
					break;
				case 0xD0: // channel pressure (aftertouch), pressure
					break;
				case 0xE0: // pitch bend, lsb, msb
					pitch_bend(channel, (data0 | (data1 << 7)) - 0x2000);
					break;
			}
		}
		
		while (synth_envelope_ticks_pending > 0 && midi_queue_count == 0) {
			for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
				synth_voice *voice = &synth_voices[v];
				synth_patch *p = &voice->patch;
				for (uint8_t o = 0; o < SYNTH_OPERATOR_COUNT; ++o) {
					synth_tick_envelope(&p->operators[o].envelope);
				}
				if (voice->velocity == 127) {
					p->operators[0].amplitude = p->operators[0].envelope.value >> (SYNTH_VOICE_COUNT / 4);
				} else {
					p->operators[0].amplitude = (p->operators[0].envelope.value >> (7 + (SYNTH_VOICE_COUNT / 4))) * (voice->velocity + 1);
				}
				if (p->operators[0].envelope.state == SYNTH_ENVELOPE_STATE_IDLE && p->operators[1].envelope.state == SYNTH_ENVELOPE_STATE_IDLE) {
					voice->state = SYNTH_VOICE_STATE_IDLE;
				}
			}
			--synth_envelope_ticks_pending;
		}
		
		
	}
}