#include <stdint.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "synth.h"
#include "waveforms.h"
#include "notes.h"

synth_voice synth_voices[SYNTH_VOICE_COUNT];
synth_patch synth_patches[SYNTH_CHANNEL_COUNT];
int16_t synth_pitch_bend[SYNTH_CHANNEL_COUNT];

volatile int8_t synth_output_a;
volatile int8_t synth_output_b;

void synth_tick_envelope(synth_envelope *envelope) {
	switch (envelope->state) {
		case SYNTH_ENVELOPE_STATE_IDLE:
			envelope->value = 0;
			break;
		case SYNTH_ENVELOPE_STATE_ATTACK:
			if (envelope->attack >= ((uint16_t)(0xFFFF - envelope->value))) {
				envelope->value = 0xFFFF;
				envelope->state = SYNTH_ENVELOPE_STATE_DECAY;
			} else {
				envelope->value += envelope->attack;
			}
			break;
		case SYNTH_ENVELOPE_STATE_DECAY:
			if (envelope->value > envelope->decay) {
				envelope->value -= envelope->decay;
			} else {
				envelope->value = envelope->sustain;
			}
			if (envelope->value <= envelope->sustain) {
				envelope->value = envelope->sustain;
				envelope->state = SYNTH_ENVELOPE_STATE_SUSTAIN;
			}
			break;
		case SYNTH_ENVELOPE_STATE_SUSTAIN:
			envelope->value = envelope->sustain;
			break;
		case SYNTH_ENVELOPE_STATE_RELEASE:
			if (envelope->release >= envelope->value) {
				envelope->value = 0;
				envelope->state = SYNTH_ENVELOPE_STATE_IDLE;
			} else if (envelope->value > envelope->sustain && envelope->value > envelope->decay) {
				envelope->value -= envelope->decay;
			} else {
				envelope->value -= envelope->release;
			}
			break;
	}
}

void synth_set_patch(uint8_t channel, synth_patch *patch) {
	synth_patches[channel] = *patch;
}

void synth_set_operator_frequency(synth_operator *op, uint8_t note, int16_t pitch_bend) {
	
	uint16_t phase_increment = pgm_read_dword(&notes[note]);
	
	uint32_t scale_factor = op->frequency_scale;
	
	pitch_bend /= 8;

	if (pitch_bend > 0) {
		scale_factor *= (pitch_bend + 0x2000);
		scale_factor >>= 13;
	} else if (pitch_bend < 0) {
		scale_factor <<= 13;
		scale_factor /= (0x2000 - pitch_bend);
	}
	
	phase_increment = (phase_increment * scale_factor) >> 10;
	
	cli();
	op->oscillator.phase_delta = phase_increment;
	sei();
	
}

void synth_set_pitch_bend(uint8_t channel, int16_t amount) {
	synth_pitch_bend[channel] = amount;
	for (uint8_t v = 0; v < SYNTH_VOICE_COUNT; ++v) {
		synth_voice *voice = &synth_voices[v];
		if (voice->channel == channel && voice->state != SYNTH_VOICE_STATE_IDLE) {
			for (uint8_t o = 0; o < SYNTH_OPERATOR_COUNT; ++o) {
				synth_operator* op = &voice->patch.operators[o];
				synth_set_operator_frequency(op, voice->note, amount);
			}
		}
	}
}

void synth_begin_voice(synth_voice *voice, uint8_t channel, uint8_t note, uint8_t velocity) {
	
	// Mark the voice as active.
	voice->state = SYNTH_VOICE_STATE_ACTIVE;
	
	// Copy the patch from the channel to the voice
	voice->patch = synth_patches[channel];
	
	// Normalise the velocity.
	uint16_t carrier_velocity = (uint16_t)velocity * (uint16_t)340;
	if (carrier_velocity > 0x7FFF) {
		carrier_velocity = 0x7F;
	} else {
		carrier_velocity = carrier_velocity >> 8;
	}
	
	// Keep track of what this voice is being used for.
	voice->channel = channel;
	voice->note = note;
	voice->velocity = carrier_velocity;
	
	
	for (uint8_t o = 0; o < SYNTH_OPERATOR_COUNT; ++o) {
		synth_operator* op = &voice->patch.operators[o];
		synth_set_operator_frequency(op, note, synth_pitch_bend[channel]);
		op->envelope.state = SYNTH_ENVELOPE_STATE_ATTACK;
	}
	
	// Carrier amplitude isn't used normally, so we'll use that to keep track of the velocity-scaled amplitude.
	voice->patch.operators[0].amplitude = 0;
	
	if (velocity < 127) {
		voice->patch.operators[1].amplitude = ((uint32_t)voice->patch.operators[1].amplitude * (uint32_t)(velocity + 1)) >> 7;
	}
}

void synth_end_voice(synth_voice *voice) {
	for (uint8_t o = 0; o < SYNTH_OPERATOR_COUNT; ++o) {
		synth_operator* op = &voice->patch.operators[o];
		if (op->envelope.state != SYNTH_ENVELOPE_STATE_IDLE) {
			op->envelope.state = SYNTH_ENVELOPE_STATE_RELEASE;
		}
	}
	voice->state = SYNTH_VOICE_STATE_ENDING;
}


volatile uint8_t envelope_advance;
volatile uint8_t synth_envelope_ticks_pending;
